#include "./ili9163/ili9163lcd.h"
#include "./spi/spi.h"
#include "./ili9163/colors.h"
#include <avr/io.h>
#include <util/delay.h>
//#include <stdlib.h>
#include <avr/interrupt.h>

DigitalPin pinSCK(&PORTB, PB7), pinSDA(&PORTB, PB5), pinA0(&PORTD, PD2), pinRESET(&PORTB, PB4 ), pinCS(&PORTD, PD4);
ILI9163 lcd (&pinA0, &pinRESET, &pinCS);
//Define directions
#define UP 0
#define LEFT 1
#define DOWN 2
#define RIGHT 3

//Define matrix elements
#define SNAKE 2
#define WALL 1
#define FREE 0
#define POINT 3

//Define states
#define MENIU 10
#define GAME 12
#define END 14

//dimensiunea marticii
#define SIZE 24

//lungimea maxima a sarpelui
#define MAX_LEN 12
int score = 0;

//pozitia sarpelui
int snake_poz_x[MAX_LEN];
int snake_poz_y[MAX_LEN];

//pozitia punctelor castigatoare
int xPoz[10]={13, 19, 10, 22, 15, 12, 20, 7, 17, 3};
int yPoz[10]={10, 17, 2, 16, 4, 15, 6, 9, 13, 5};
int cont = 0;

int matrix[SIZE][SIZE];

//dimensiunea sarpelui
int snakeSize = 2;

//directia de miscare
int direction = UP;

//stare jocului (MENIU, GAME, sau END)
int state = MENIU;

int select = 1; //linia din meniu selectata
int level = 1;  //nivelul jocului
int level_selected = 0; //nivelul selectat in meniu
int speed = 3;          //viteza jocului
int speed_selected = 0; //viteza selectata in meniu
int level_w, speed_w;
int done = 0;
void init_game() {
	done = 0;
	score = 0;
	snakeSize = 2;
	direction = UP;
	for (int i = 0; i < MAX_LEN; i++) {
		snake_poz_x[i] = 0;
		snake_poz_y[i] = 0;
	}
	snake_poz_x[0] = 15;
	snake_poz_y[0] = 13;
	snake_poz_x[1] = 16;
	snake_poz_y[1] = 13;
}
void init_matrix_level1(){

	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			if (i == 0 || i == SIZE - 1 || j == 0 || j == SIZE - 1) {
				matrix[i][j] = WALL;
			}
			else if((i == 15 || i == 16) && j == 13){
				matrix[i][j] = SNAKE;
				
			}else
				matrix[i][j] = FREE;
		}
	}
}
void init_matrix_level2() {

	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			if (i == 0 || i == SIZE - 1 || j == 0 || j == SIZE - 1) {
				matrix[i][j] = WALL;
			}
			else if ((i == 15 || i == 16) && j == 13) {
				matrix[i][j] = SNAKE;
			}
			else if (i == 14 && j <= 10 || i == 7 && j > 13) {
				matrix[i][j] = WALL;
			}
			else
				matrix[i][j] = FREE;
		}
	}
}
void init_matrix_level3() {

	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			if (i == 0 || i == SIZE - 1 || j == 0 || j == SIZE - 1) {
				matrix[i][j] = WALL;
			}
			else if ((i == 15 || i == 16) && j == 13) {
				matrix[i][j] = SNAKE;
			}
			else if ((i <= 6  && j == 7) 
				|| (i == 6 && j >=7 && j <=9 ) || (i <= 6  && j == 14)
				|| (i == 6 && j >= 14 && j <= 19) || (j == 19 && i >=7 && i <= 9)
				|| (i == 14 && j <= 5) || (j == 5 && i >= 14 && i <= 17)
				|| (i == 16 && j>=16 ) || ( j == 16 && i <= 16 && i>= 13)
				|| (j == 10 && i>= 18)) {
				matrix[i][j] = WALL;
			}
			else
				matrix[i][j] = FREE;
		}
	}
}
void print_matrix(){
	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			if (matrix[i][j] == 1) {
				lcd.drawPixel(10 * j, 10*i, 10, BLUE);
				
			}else if (matrix[i][j] == 2){
				lcd.drawPixel(10 * j, 10*i, 10, GREEN);
			}
		}
	}
}

void sound1(){
	//castigarea unui punct
	PORTD &= ~(1 << PD4);
	_delay_ms(100);
	PORTD |= (1 << PD4);
	_delay_ms(100);
	PORTD &= ~(1 << PD4);
	_delay_ms(100);
	PORTD |= (1 << PD4);
}
void sound2(){
	//pierederea jocului
	PORTD &= ~(1 << PD4);
	_delay_ms(500);
	PORTD |= (1 << PD4);
}

//generarea unui punct castigator
void generate_points(){
	int x = xPoz[cont];
	int y = yPoz[cont];
	matrix[x][y] = POINT;
	lcd.drawCircle(10*y+5, 10*x+5, 5, RED);
	if (cont < 9)
		cont++;
	else
		cont = 0;
}

//mutarea sarpelui cu o pozitie in directia data de 'direction'
//returneaza 0 la mutarea cu succes si -1 daca nu se poate muta
int move_snake(){
	int headX = snake_poz_x[0];
	int headY = snake_poz_y[0];

	int x, y;
	if (direction == UP) {
		x = -1;
		y = 0;
	}
	else if (direction == LEFT) {
		x = 0;
		y = -1;
	}
	else if (direction == DOWN) {
		x = 1;
		y = 0;
		
	}else if (direction == RIGHT) {
		x = 0;
		y = 1;
	}
	
	//daca dau peste un perete sau un segment al sarpelui
	if(matrix[headX + x][headY + y] == WALL || matrix[headX + x][headY + y] == SNAKE){
		sound2();
		return -1;
	}
	//daca dau peste un punct castigator (lungimea sarpelui creste)
	else if(matrix[headX + x][headY + y] == POINT)
	{
		matrix[headX + x][headY + y] = SNAKE;
		lcd.drawCircle(10*(headY + y)+5, 10*(headX + x)+5, 5, BLACK);
		lcd.drawPixel(10 * (headY + y), 10 * (headX + x), 10, GREEN);
		
		for (int i = snakeSize ; i > 0; i--) {
			snake_poz_x[i] = snake_poz_x[i - 1];
			snake_poz_y[i] = snake_poz_y[i - 1];
		}
		snake_poz_x[0] = headX + x;
		snake_poz_y[0] = headY + y;
		
		score++;
		snakeSize++;
		sound1();
		if(snakeSize < MAX_LEN)
			generate_points();
		else {
			return -1;
		}
	}
	//daca pozitia este libera
	else if(matrix[headX + x][headY + y] == FREE){
		matrix[headX + x][headY + y] = SNAKE;
		lcd.drawPixel(10 * (headY + y), 10*(headX + x), 10, GREEN);
		
		matrix[snake_poz_x[snakeSize - 1]][snake_poz_y[snakeSize - 1]] = FREE;
		lcd.drawPixel(10 * snake_poz_y[snakeSize - 1], 10 * snake_poz_x[snakeSize - 1], 10, BLACK);
		
		for (int i = snakeSize ; i > 0; i--) {
			snake_poz_x[i] = snake_poz_x[i - 1];
			snake_poz_y[i] = snake_poz_y[i - 1];
		}
		snake_poz_x[0] = headX + x;
		snake_poz_y[0] = headY + y;
		snake_poz_x[snakeSize] = 0;
		snake_poz_y[snakeSize] = 0;

	}
	return 0;
}

//initializarea meniului
void init_meniu(){
	lcd.drawClear(BLACK);
	lcd.drawString(80, 30, 3, RED, TRANSPARENT, "SNAKE");
	lcd.drawString(80, 70, 3, YELLOW, TRANSPARENT, "MENIU");
	
	lcd.drawString(10, 120, 2, GREEN, WHITE, "* Play Game");
	lcd.drawString(10, 150, 2, GREEN, TRANSPARENT, "* Select Level");
	lcd.drawString(10, 180, 2, GREEN, TRANSPARENT, "* Select Speed");
}

//tratarea intreruperilor pentru butoane
ISR(PCINT1_vect)
{
	if (state == GAME) {
		if ((PINB & (1 << PB0)) == 0 && (direction == LEFT || direction == RIGHT)) {
			direction = DOWN;
		}
		if ((PINB & (1 << PB1)) == 0 && (direction == UP || direction == DOWN)) {
			direction = LEFT;
		}
		if ((PINB & (1 << PB2)) == 0 && (direction == LEFT || direction == RIGHT)) {
			direction = UP;
		}

		if ((PINB & (1 << PB3)) == 0 && (direction == UP || direction == DOWN)) {
			direction = RIGHT;
		}
	}
}

//intreruperea la overflow a timerului 0
//in functie de viteza se muta sarpele dupa un anumit numar de intreruperi
volatile int total_ovf = 0;
ISR(TIMER0_OVF_vect)
{
	total_ovf++;
	int t;
	if (speed == 1)
		t = 30;
	else if (speed == 2)
		t = 15;
	else if (speed == 3)
		t = 9;

	if (total_ovf == t) {
		
		total_ovf = 0;
		if (state == GAME) {
			if (move_snake() == -1) {
				state = END;
			}
		}
	}
	
}

void IO_init() {

	//iesire buzzer
	DDRD |= (1 << PD4);
	//opresc buzzerul
	PORTD |= (1 << PD4);

	//intrare (butoane)
	DDRB &= ~(1 << PB0);
	PORTB |= (1 << PB0);
	DDRB &= ~(1 << PB1);
	PORTB |= (1 << PB1);
	DDRB &= ~(1 << PB2);
	PORTB |= (1 << PB2);
	DDRB &= ~(1 << PB3);
	PORTB |= (1 << PB3);

	// Activez PIN CHANGE INTERRUPT ENABLE pentru PCINT15:8 (portul B)
	PCICR |= (1 << PCIE1);

	//Activez intreruperea pe butoanele PB0 PB1 PB2 PB3
	PCMSK1 |= (1 << PCINT8 ); //PB0
	PCMSK1 |= (1 << PCINT9);  //PB1
	PCMSK1 |= (1 << PCINT10); //PB2
	PCMSK1 |= (1 << PCINT11); //PB3
}

//initializarea timerului 0
void timer0_init() {
	TCCR0A = 0;
	TCCR0B |= (1 << CS02) | (1 << CS00);
	TCNT0 = 0;
	TIMSK0 |= (1 << TOIE0);
}
int main(void){
	//initializari
	spi_init();
	lcd.init(ILI9163::ROT0);
	sei();
	IO_init();
	lcd.drawClear(BLACK);
	init_meniu();
//----------------------------------------------------------------------------------------
	while(1){
		if (state == MENIU) {
			if ((PINB & (1 << PB0)) == 0 && select == 1){
				//buton down
				lcd.drawString(10, 120, 2, GREEN, BLACK, "* Play Game");
				lcd.drawString(10, 150, 2, GREEN, WHITE, "* Select Level");
				select = 2;
				_delay_ms(100);
			}
			if ((PINB & (1 << PB3)) == 0 && select == 1){
				//buton right
				_delay_ms(100);
				if(level == 1)
					init_matrix_level1();
				else if(level == 2)
					init_matrix_level2();
				else if(level == 3)
					init_matrix_level3();
				timer0_init();
				init_game();
				lcd.drawClear(BLACK);
				print_matrix();
				generate_points();
				state = GAME;
			}
			if ((PINB & (1 << PB2)) == 0 && select == 2&& level_selected == 0){
				//buton up
				lcd.drawString(10, 120, 2, GREEN, WHITE, "* Play Game");
				lcd.drawString(10, 150, 2, GREEN, BLACK, "* Select Level");
				select = 1;
				_delay_ms(100);
			}
			if ((PINB & (1 << PB0)) == 0 && select == 2&& level_selected == 0){
				//buton down
				lcd.drawString(10, 150, 2, GREEN, BLACK, "* Select Level");
				lcd.drawString(10, 180, 2, GREEN, WHITE, "* Select Speed");
				select = 3;
				_delay_ms(100);
			}
//====================================== select level =====================================
			if((PINB & (1 << PB3)) == 0 && select == 2 && level_selected == 0){
				//buton right
				lcd.drawString(200, 150, 2, RED, BLACK, "1");
				lcd.drawString(200, 180, 2, RED, BLACK, "2");
				lcd.drawString(200, 210, 2, RED, BLACK, "3");
				
				if(level == 1){
					level_w = 150;
				}else if(level == 2){
					level_w = 180;
				
				}else if(level == 3){
					level_w = 210;
				}
				lcd.drawString(210, level_w, 2, RED, BLACK, "<-");
				level_selected = level;
				_delay_ms(100);
			}
			if ((PINB & (1 << PB0)) == 0 && select == 2 && level_selected != 0) {
				//buton down

				lcd.drawString(210, level_w, 2, BLACK, BLACK, "<-");
				if (level_w == 210) level_w = 150;
				else level_w += 30;
				lcd.drawString(210, level_w, 2, RED, BLACK, "<-");

				if (level == 3) level = 1;
				else level++;

				level_selected = level;

				_delay_ms(100);

			}
			if ((PINB & (1 << PB2)) == 0 && select == 2 && level_selected != 0) {
				//buton up
				lcd.drawString(210, level_w, 2, BLACK, BLACK, "<-");
				if (level_w == 150) level_w = 210;
				else level_w -= 30;
				lcd.drawString(210, level_w, 2, RED, BLACK, "<-");

				if (level == 1) level = 3;
				else level--;
				level_selected = level;
				_delay_ms(100);
			}
			if ((PINB & (1 << PB3)) == 0 && select == 2 && level_selected != 0) {
				//buton right
				lcd.drawString(200, 150, 2, BLACK, BLACK, "1");
				lcd.drawString(200, 180, 2, BLACK, BLACK, "2");
				lcd.drawString(200, 210, 2, BLACK, BLACK, "3");
				lcd.drawString(210, level_w, 2, BLACK, BLACK, "<-");
				level = level_selected;
				_delay_ms(100);
				level_selected = 0;
			}
	
			if ((PINB & (1 << PB2)) == 0 && select == 3 && speed_selected == 0){ 
				//buton up
				lcd.drawString(10, 150, 2, GREEN, WHITE, "* Select Level");
				lcd.drawString(10, 180, 2, GREEN, BLACK, "* Select Speed");
				select = 2;
				_delay_ms(100);
			}
//=========================================== select speed =================================	
			if ((PINB & (1 << PB3)) == 0 && select == 3 && speed_selected == 0) {
				//buton right
				lcd.drawString(200, 180, 2, RED, BLACK, "1");
				lcd.drawString(200, 210, 2, RED, BLACK, "2");
				lcd.drawString(200, 240, 2, RED, BLACK, "3");

				if (speed == 1) speed_w = 180;
				else if (speed == 2) speed_w = 210;
				else if (speed == 3) speed_w = 240;

				lcd.drawString(210, speed_w, 2, RED, BLACK, "<-");

				speed_selected = speed;
				_delay_ms(100);
			}
			if ((PINB & (1 << PB0)) == 0 && select == 3 && speed_selected != 0 ) {
				//buton down
				lcd.drawString(210, speed_w, 2, BLACK, BLACK, "<-");
				if (speed_w == 240) speed_w = 180;
				else speed_w += 30;
				lcd.drawString(210, speed_w, 2, RED, BLACK, "<-");

				if (speed == 3) speed = 1;
				else speed++;

				speed_selected = speed;
				_delay_ms(100);
			}
			if ((PINB & (1 << PB2)) == 0 && select == 3 && speed_selected != 0) {
				//buton up

				lcd.drawString(210, speed_w, 2, BLACK, BLACK, "<-");
				if (speed_w == 180) speed_w = 240;
				else speed_w -= 30;
				lcd.drawString(210, speed_w, 2, RED, BLACK, "<-");

				if (speed == 1) speed = 3;
				else speed--;

				speed_selected = speed;
				_delay_ms(100);
			}
			if ((PINB & (1 << PB3)) == 0 && select == 3 && speed_selected != 0) {
				//buton right
				lcd.drawString(200, 180, 2, BLACK, BLACK, "1");
				lcd.drawString(200, 210, 2, BLACK, BLACK, "2");
				lcd.drawString(200, 240, 2, BLACK, BLACK, "3");
				lcd.drawString(210, speed_w, 2, BLACK, BLACK, "<-");
				speed = speed_selected;
				_delay_ms(100);
				speed_selected = 0;
			}
		}
		else if (state == END) {
			if (!done) {
				lcd.drawClear(BLACK);
				if (snakeSize == MAX_LEN) {
					lcd.drawStringF(15, 50, 3, RED, TRANSPARENT, "You won! :)");
					lcd.drawStringF(60, 130, 2, YELLOW, TRANSPARENT, "Score: %d", score);
					if (level != 3)
						lcd.drawString(20, 160, 2, YELLOW, TRANSPARENT, "Play next level?");
					else 
						lcd.drawString(40, 160, 2, YELLOW, TRANSPARENT, "Play again?");
					
					lcd.drawString(90, 190, 2, MAGENTA, TRANSPARENT, "Yes ->");
					lcd.drawString(50, 220, 2, MAGENTA, TRANSPARENT, "<- No");
				}
				else {
					lcd.drawStringF(15, 50, 3, RED, TRANSPARENT, "Try again :(");
					lcd.drawStringF(60, 130, 2, YELLOW, TRANSPARENT, "Score: %d", score);
					lcd.drawString(40, 160, 2, YELLOW, TRANSPARENT, "Play again?");
					lcd.drawString(90, 190, 2, MAGENTA, TRANSPARENT, "Yes ->");
					lcd.drawString(50, 220, 2, MAGENTA, TRANSPARENT, "<- No");
				}
				done = 1;
			}

			if ((PINB & (1 << PB3)) == 0 && snakeSize != MAX_LEN) {
				//buton right
				init_game();
				
				if (level == 1) init_matrix_level1();
				else if (level == 2) init_matrix_level2();
				else if (level == 3) init_matrix_level3();

				lcd.drawClear(BLACK);
				print_matrix();
				generate_points();
				state = GAME;
			}
			if ((PINB & (1 << PB3)) == 0 && snakeSize == MAX_LEN) {
				//buton right
				init_game();
				if (level == 1 || level == 2)
					level++;

				if (level == 1) init_matrix_level1();
				else if (level == 2) init_matrix_level2();
				else if (level == 3) init_matrix_level3();

				lcd.drawClear(BLACK);
				print_matrix();
				generate_points();
				state = GAME;
			}

			if ((PINB & (1 << PB1)) == 0) {
				//buton left
				state = MENIU;
				init_game();
				lcd.drawClear(BLACK);
				init_meniu();
			}
		}
	}
	return 0;
}